# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import btrdb.grpcinterface.btrdb_pb2 as btrdb__pb2


class BTrDBStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RawValues = channel.unary_stream(
                '/v5api.BTrDB/RawValues',
                request_serializer=btrdb__pb2.RawValuesParams.SerializeToString,
                response_deserializer=btrdb__pb2.RawValuesResponse.FromString,
                )
        self.ArrowRawValues = channel.unary_stream(
                '/v5api.BTrDB/ArrowRawValues',
                request_serializer=btrdb__pb2.ArrowRawValuesParams.SerializeToString,
                response_deserializer=btrdb__pb2.ArrowRawValuesResponse.FromString,
                )
        self.ArrowMultiValues = channel.unary_stream(
                '/v5api.BTrDB/ArrowMultiValues',
                request_serializer=btrdb__pb2.ArrowMultiValuesParams.SerializeToString,
                response_deserializer=btrdb__pb2.ArrowMultiValuesResponse.FromString,
                )
        self.AlignedWindows = channel.unary_stream(
                '/v5api.BTrDB/AlignedWindows',
                request_serializer=btrdb__pb2.AlignedWindowsParams.SerializeToString,
                response_deserializer=btrdb__pb2.AlignedWindowsResponse.FromString,
                )
        self.ArrowAlignedWindows = channel.unary_stream(
                '/v5api.BTrDB/ArrowAlignedWindows',
                request_serializer=btrdb__pb2.AlignedWindowsParams.SerializeToString,
                response_deserializer=btrdb__pb2.ArrowAlignedWindowsResponse.FromString,
                )
        self.Windows = channel.unary_stream(
                '/v5api.BTrDB/Windows',
                request_serializer=btrdb__pb2.WindowsParams.SerializeToString,
                response_deserializer=btrdb__pb2.WindowsResponse.FromString,
                )
        self.ArrowWindows = channel.unary_stream(
                '/v5api.BTrDB/ArrowWindows',
                request_serializer=btrdb__pb2.WindowsParams.SerializeToString,
                response_deserializer=btrdb__pb2.ArrowWindowsResponse.FromString,
                )
        self.StreamInfo = channel.unary_unary(
                '/v5api.BTrDB/StreamInfo',
                request_serializer=btrdb__pb2.StreamInfoParams.SerializeToString,
                response_deserializer=btrdb__pb2.StreamInfoResponse.FromString,
                )
        self.SetStreamAnnotations = channel.unary_unary(
                '/v5api.BTrDB/SetStreamAnnotations',
                request_serializer=btrdb__pb2.SetStreamAnnotationsParams.SerializeToString,
                response_deserializer=btrdb__pb2.SetStreamAnnotationsResponse.FromString,
                )
        self.SetStreamTags = channel.unary_unary(
                '/v5api.BTrDB/SetStreamTags',
                request_serializer=btrdb__pb2.SetStreamTagsParams.SerializeToString,
                response_deserializer=btrdb__pb2.SetStreamTagsResponse.FromString,
                )
        self.Create = channel.unary_unary(
                '/v5api.BTrDB/Create',
                request_serializer=btrdb__pb2.CreateParams.SerializeToString,
                response_deserializer=btrdb__pb2.CreateResponse.FromString,
                )
        self.ListCollections = channel.unary_stream(
                '/v5api.BTrDB/ListCollections',
                request_serializer=btrdb__pb2.ListCollectionsParams.SerializeToString,
                response_deserializer=btrdb__pb2.ListCollectionsResponse.FromString,
                )
        self.LookupStreams = channel.unary_stream(
                '/v5api.BTrDB/LookupStreams',
                request_serializer=btrdb__pb2.LookupStreamsParams.SerializeToString,
                response_deserializer=btrdb__pb2.LookupStreamsResponse.FromString,
                )
        self.Nearest = channel.unary_unary(
                '/v5api.BTrDB/Nearest',
                request_serializer=btrdb__pb2.NearestParams.SerializeToString,
                response_deserializer=btrdb__pb2.NearestResponse.FromString,
                )
        self.Changes = channel.unary_stream(
                '/v5api.BTrDB/Changes',
                request_serializer=btrdb__pb2.ChangesParams.SerializeToString,
                response_deserializer=btrdb__pb2.ChangesResponse.FromString,
                )
        self.Insert = channel.unary_unary(
                '/v5api.BTrDB/Insert',
                request_serializer=btrdb__pb2.InsertParams.SerializeToString,
                response_deserializer=btrdb__pb2.InsertResponse.FromString,
                )
        self.ArrowInsert = channel.unary_unary(
                '/v5api.BTrDB/ArrowInsert',
                request_serializer=btrdb__pb2.ArrowInsertParams.SerializeToString,
                response_deserializer=btrdb__pb2.InsertResponse.FromString,
                )
        self.Delete = channel.unary_unary(
                '/v5api.BTrDB/Delete',
                request_serializer=btrdb__pb2.DeleteParams.SerializeToString,
                response_deserializer=btrdb__pb2.DeleteResponse.FromString,
                )
        self.Info = channel.unary_unary(
                '/v5api.BTrDB/Info',
                request_serializer=btrdb__pb2.InfoParams.SerializeToString,
                response_deserializer=btrdb__pb2.InfoResponse.FromString,
                )
        self.FaultInject = channel.unary_unary(
                '/v5api.BTrDB/FaultInject',
                request_serializer=btrdb__pb2.FaultInjectParams.SerializeToString,
                response_deserializer=btrdb__pb2.FaultInjectResponse.FromString,
                )
        self.Flush = channel.unary_unary(
                '/v5api.BTrDB/Flush',
                request_serializer=btrdb__pb2.FlushParams.SerializeToString,
                response_deserializer=btrdb__pb2.FlushResponse.FromString,
                )
        self.Obliterate = channel.unary_unary(
                '/v5api.BTrDB/Obliterate',
                request_serializer=btrdb__pb2.ObliterateParams.SerializeToString,
                response_deserializer=btrdb__pb2.ObliterateResponse.FromString,
                )
        self.GetMetadataUsage = channel.unary_unary(
                '/v5api.BTrDB/GetMetadataUsage',
                request_serializer=btrdb__pb2.MetadataUsageParams.SerializeToString,
                response_deserializer=btrdb__pb2.MetadataUsageResponse.FromString,
                )
        self.GenerateCSV = channel.unary_stream(
                '/v5api.BTrDB/GenerateCSV',
                request_serializer=btrdb__pb2.GenerateCSVParams.SerializeToString,
                response_deserializer=btrdb__pb2.GenerateCSVResponse.FromString,
                )
        self.SQLQuery = channel.unary_stream(
                '/v5api.BTrDB/SQLQuery',
                request_serializer=btrdb__pb2.SQLQueryParams.SerializeToString,
                response_deserializer=btrdb__pb2.SQLQueryResponse.FromString,
                )
        self.Subscribe = channel.stream_stream(
                '/v5api.BTrDB/Subscribe',
                request_serializer=btrdb__pb2.SubscriptionUpdate.SerializeToString,
                response_deserializer=btrdb__pb2.SubscriptionResp.FromString,
                )


class BTrDBServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RawValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArrowRawValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArrowMultiValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AlignedWindows(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArrowAlignedWindows(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Windows(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArrowWindows(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetStreamAnnotations(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetStreamTags(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Create(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListCollections(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LookupStreams(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Nearest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Changes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Insert(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArrowInsert(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Delete(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Info(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FaultInject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Flush(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Obliterate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMetadataUsage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GenerateCSV(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SQLQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request_iterator, context):
        """rpc SetCompactionConfig(SetCompactionConfigParams) returns (SetCompactionConfigResponse);
        rpc GetCompactionConfig(GetCompactionConfigParams) returns (GetCompactionConfigResponse);
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BTrDBServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RawValues': grpc.unary_stream_rpc_method_handler(
                    servicer.RawValues,
                    request_deserializer=btrdb__pb2.RawValuesParams.FromString,
                    response_serializer=btrdb__pb2.RawValuesResponse.SerializeToString,
            ),
            'ArrowRawValues': grpc.unary_stream_rpc_method_handler(
                    servicer.ArrowRawValues,
                    request_deserializer=btrdb__pb2.ArrowRawValuesParams.FromString,
                    response_serializer=btrdb__pb2.ArrowRawValuesResponse.SerializeToString,
            ),
            'ArrowMultiValues': grpc.unary_stream_rpc_method_handler(
                    servicer.ArrowMultiValues,
                    request_deserializer=btrdb__pb2.ArrowMultiValuesParams.FromString,
                    response_serializer=btrdb__pb2.ArrowMultiValuesResponse.SerializeToString,
            ),
            'AlignedWindows': grpc.unary_stream_rpc_method_handler(
                    servicer.AlignedWindows,
                    request_deserializer=btrdb__pb2.AlignedWindowsParams.FromString,
                    response_serializer=btrdb__pb2.AlignedWindowsResponse.SerializeToString,
            ),
            'ArrowAlignedWindows': grpc.unary_stream_rpc_method_handler(
                    servicer.ArrowAlignedWindows,
                    request_deserializer=btrdb__pb2.AlignedWindowsParams.FromString,
                    response_serializer=btrdb__pb2.ArrowAlignedWindowsResponse.SerializeToString,
            ),
            'Windows': grpc.unary_stream_rpc_method_handler(
                    servicer.Windows,
                    request_deserializer=btrdb__pb2.WindowsParams.FromString,
                    response_serializer=btrdb__pb2.WindowsResponse.SerializeToString,
            ),
            'ArrowWindows': grpc.unary_stream_rpc_method_handler(
                    servicer.ArrowWindows,
                    request_deserializer=btrdb__pb2.WindowsParams.FromString,
                    response_serializer=btrdb__pb2.ArrowWindowsResponse.SerializeToString,
            ),
            'StreamInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.StreamInfo,
                    request_deserializer=btrdb__pb2.StreamInfoParams.FromString,
                    response_serializer=btrdb__pb2.StreamInfoResponse.SerializeToString,
            ),
            'SetStreamAnnotations': grpc.unary_unary_rpc_method_handler(
                    servicer.SetStreamAnnotations,
                    request_deserializer=btrdb__pb2.SetStreamAnnotationsParams.FromString,
                    response_serializer=btrdb__pb2.SetStreamAnnotationsResponse.SerializeToString,
            ),
            'SetStreamTags': grpc.unary_unary_rpc_method_handler(
                    servicer.SetStreamTags,
                    request_deserializer=btrdb__pb2.SetStreamTagsParams.FromString,
                    response_serializer=btrdb__pb2.SetStreamTagsResponse.SerializeToString,
            ),
            'Create': grpc.unary_unary_rpc_method_handler(
                    servicer.Create,
                    request_deserializer=btrdb__pb2.CreateParams.FromString,
                    response_serializer=btrdb__pb2.CreateResponse.SerializeToString,
            ),
            'ListCollections': grpc.unary_stream_rpc_method_handler(
                    servicer.ListCollections,
                    request_deserializer=btrdb__pb2.ListCollectionsParams.FromString,
                    response_serializer=btrdb__pb2.ListCollectionsResponse.SerializeToString,
            ),
            'LookupStreams': grpc.unary_stream_rpc_method_handler(
                    servicer.LookupStreams,
                    request_deserializer=btrdb__pb2.LookupStreamsParams.FromString,
                    response_serializer=btrdb__pb2.LookupStreamsResponse.SerializeToString,
            ),
            'Nearest': grpc.unary_unary_rpc_method_handler(
                    servicer.Nearest,
                    request_deserializer=btrdb__pb2.NearestParams.FromString,
                    response_serializer=btrdb__pb2.NearestResponse.SerializeToString,
            ),
            'Changes': grpc.unary_stream_rpc_method_handler(
                    servicer.Changes,
                    request_deserializer=btrdb__pb2.ChangesParams.FromString,
                    response_serializer=btrdb__pb2.ChangesResponse.SerializeToString,
            ),
            'Insert': grpc.unary_unary_rpc_method_handler(
                    servicer.Insert,
                    request_deserializer=btrdb__pb2.InsertParams.FromString,
                    response_serializer=btrdb__pb2.InsertResponse.SerializeToString,
            ),
            'ArrowInsert': grpc.unary_unary_rpc_method_handler(
                    servicer.ArrowInsert,
                    request_deserializer=btrdb__pb2.ArrowInsertParams.FromString,
                    response_serializer=btrdb__pb2.InsertResponse.SerializeToString,
            ),
            'Delete': grpc.unary_unary_rpc_method_handler(
                    servicer.Delete,
                    request_deserializer=btrdb__pb2.DeleteParams.FromString,
                    response_serializer=btrdb__pb2.DeleteResponse.SerializeToString,
            ),
            'Info': grpc.unary_unary_rpc_method_handler(
                    servicer.Info,
                    request_deserializer=btrdb__pb2.InfoParams.FromString,
                    response_serializer=btrdb__pb2.InfoResponse.SerializeToString,
            ),
            'FaultInject': grpc.unary_unary_rpc_method_handler(
                    servicer.FaultInject,
                    request_deserializer=btrdb__pb2.FaultInjectParams.FromString,
                    response_serializer=btrdb__pb2.FaultInjectResponse.SerializeToString,
            ),
            'Flush': grpc.unary_unary_rpc_method_handler(
                    servicer.Flush,
                    request_deserializer=btrdb__pb2.FlushParams.FromString,
                    response_serializer=btrdb__pb2.FlushResponse.SerializeToString,
            ),
            'Obliterate': grpc.unary_unary_rpc_method_handler(
                    servicer.Obliterate,
                    request_deserializer=btrdb__pb2.ObliterateParams.FromString,
                    response_serializer=btrdb__pb2.ObliterateResponse.SerializeToString,
            ),
            'GetMetadataUsage': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMetadataUsage,
                    request_deserializer=btrdb__pb2.MetadataUsageParams.FromString,
                    response_serializer=btrdb__pb2.MetadataUsageResponse.SerializeToString,
            ),
            'GenerateCSV': grpc.unary_stream_rpc_method_handler(
                    servicer.GenerateCSV,
                    request_deserializer=btrdb__pb2.GenerateCSVParams.FromString,
                    response_serializer=btrdb__pb2.GenerateCSVResponse.SerializeToString,
            ),
            'SQLQuery': grpc.unary_stream_rpc_method_handler(
                    servicer.SQLQuery,
                    request_deserializer=btrdb__pb2.SQLQueryParams.FromString,
                    response_serializer=btrdb__pb2.SQLQueryResponse.SerializeToString,
            ),
            'Subscribe': grpc.stream_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=btrdb__pb2.SubscriptionUpdate.FromString,
                    response_serializer=btrdb__pb2.SubscriptionResp.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'v5api.BTrDB', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class BTrDB(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RawValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/RawValues',
            btrdb__pb2.RawValuesParams.SerializeToString,
            btrdb__pb2.RawValuesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArrowRawValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/ArrowRawValues',
            btrdb__pb2.ArrowRawValuesParams.SerializeToString,
            btrdb__pb2.ArrowRawValuesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArrowMultiValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/ArrowMultiValues',
            btrdb__pb2.ArrowMultiValuesParams.SerializeToString,
            btrdb__pb2.ArrowMultiValuesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AlignedWindows(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/AlignedWindows',
            btrdb__pb2.AlignedWindowsParams.SerializeToString,
            btrdb__pb2.AlignedWindowsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArrowAlignedWindows(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/ArrowAlignedWindows',
            btrdb__pb2.AlignedWindowsParams.SerializeToString,
            btrdb__pb2.ArrowAlignedWindowsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Windows(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/Windows',
            btrdb__pb2.WindowsParams.SerializeToString,
            btrdb__pb2.WindowsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArrowWindows(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/ArrowWindows',
            btrdb__pb2.WindowsParams.SerializeToString,
            btrdb__pb2.ArrowWindowsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StreamInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/StreamInfo',
            btrdb__pb2.StreamInfoParams.SerializeToString,
            btrdb__pb2.StreamInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetStreamAnnotations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/SetStreamAnnotations',
            btrdb__pb2.SetStreamAnnotationsParams.SerializeToString,
            btrdb__pb2.SetStreamAnnotationsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetStreamTags(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/SetStreamTags',
            btrdb__pb2.SetStreamTagsParams.SerializeToString,
            btrdb__pb2.SetStreamTagsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Create(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Create',
            btrdb__pb2.CreateParams.SerializeToString,
            btrdb__pb2.CreateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListCollections(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/ListCollections',
            btrdb__pb2.ListCollectionsParams.SerializeToString,
            btrdb__pb2.ListCollectionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LookupStreams(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/LookupStreams',
            btrdb__pb2.LookupStreamsParams.SerializeToString,
            btrdb__pb2.LookupStreamsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Nearest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Nearest',
            btrdb__pb2.NearestParams.SerializeToString,
            btrdb__pb2.NearestResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Changes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/Changes',
            btrdb__pb2.ChangesParams.SerializeToString,
            btrdb__pb2.ChangesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Insert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Insert',
            btrdb__pb2.InsertParams.SerializeToString,
            btrdb__pb2.InsertResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArrowInsert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/ArrowInsert',
            btrdb__pb2.ArrowInsertParams.SerializeToString,
            btrdb__pb2.InsertResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Delete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Delete',
            btrdb__pb2.DeleteParams.SerializeToString,
            btrdb__pb2.DeleteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Info(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Info',
            btrdb__pb2.InfoParams.SerializeToString,
            btrdb__pb2.InfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FaultInject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/FaultInject',
            btrdb__pb2.FaultInjectParams.SerializeToString,
            btrdb__pb2.FaultInjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Flush(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Flush',
            btrdb__pb2.FlushParams.SerializeToString,
            btrdb__pb2.FlushResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Obliterate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/Obliterate',
            btrdb__pb2.ObliterateParams.SerializeToString,
            btrdb__pb2.ObliterateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMetadataUsage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v5api.BTrDB/GetMetadataUsage',
            btrdb__pb2.MetadataUsageParams.SerializeToString,
            btrdb__pb2.MetadataUsageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GenerateCSV(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/GenerateCSV',
            btrdb__pb2.GenerateCSVParams.SerializeToString,
            btrdb__pb2.GenerateCSVResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SQLQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/v5api.BTrDB/SQLQuery',
            btrdb__pb2.SQLQueryParams.SerializeToString,
            btrdb__pb2.SQLQueryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Subscribe(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/v5api.BTrDB/Subscribe',
            btrdb__pb2.SubscriptionUpdate.SerializeToString,
            btrdb__pb2.SubscriptionResp.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
import os
from uuid import uuid4 as new_uuid

import pytest

import btrdb
from btrdb.exceptions import BTrDBError
from btrdb.utils.credentials import credentials_by_profile


@pytest.mark.skipif(
    os.getenv("BTRDB_INTEGRATION_TEST_PROFILE") is None,
    reason="Need an integration test server to run integration tests.",
)
class TestConnection:
    def test_connection_info(self, conn):
        info = conn.info()
        logging.info(f"connection info: {info}")

    def test_incorrect_connect(
        self,
    ):
        env_name = os.getenv("BTRDB_INTEGRATION_TEST_PROFILE")
        creds = credentials_by_profile(env_name)
        conn_str = creds["endpoints"]
        err_msg = r"""Could not connect to the database, error message: <_InactiveRpcError of RPC that terminated with:\n\tstatus = StatusCode.UNAUTHENTICATED\n\tdetails = "invalid api key"\n"""
        with pytest.raises(BTrDBError, match=err_msg):
            conn = btrdb.connect(conn_str=conn_str, apikey="BOGUS_KEY")

    @pytest.mark.xfail(
        reason="Should return BTrDBError, but returns GRPCError instead, FIXME"
    )
    def test_from_uuid(self, conn):
        # todo, investigate this, the stat code returned is 0, which is why its not being returned as a btrdb error
        uu = new_uuid()
        stream = conn.stream_from_uuid(uu)
        with pytest.raises(btrdb.exceptions.BTrDBError):
            print(stream)

    def test_create_stream(self, conn, tmp_collection):
        uuid = new_uuid()
        stream = conn.create(uuid, tmp_collection, tags={"name": "s"})
        assert stream.uuid == uuid
        assert stream.name == "s"

    def test_query(self, conn, tmp_collection):
        conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
        conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
        uuids = conn.query(
            "select name from streams where collection = $1 order by name;",
            [tmp_collection],
        )
        assert len(uuids) == 2
        assert uuids[0]["name"] == "s1"
        assert uuids[1]["name"] == "s2"

    def test_list_collections(self, conn, tmp_collection):
        assert tmp_collection not in conn.list_collections()
        stream = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
        assert tmp_collection in conn.list_collections()



from uuid import uuid4 as new_uuid

import pytest


@pytest.mark.xfail
def test_create_count_obliterate_concurrency_bug(conn, tmp_collection):
    from concurrent.futures import ThreadPoolExecutor

    n_streams = 10

    def create_stream(i):
        return conn.create(new_uuid(), tmp_collection, tags={"name": f"s{i}"})

    def points_in_stream(s):
        return s.count()

    def obliterate_stream(s):
        s.obliterate()

    with ThreadPoolExecutor() as executor:
        for i in range(2):
            pmap = lambda f, it: list(executor.map(f, it, timeout=30))
            streams = pmap(create_stream, range(n_streams))
            assert sum(pmap(points_in_stream, streams)) == 0
            pmap(obliterate_stream, streams)



# btrdb.version
# Maintains version and package information for deployment.
#
# Author:   PingThings
# Created:  Tue Dec 18 14:50:05 2018 -0500
#
# For license information, see LICENSE.txt
# ID: version.py [] allen@pingthings.io $

"""
Maintains version and package information for deployment.
"""

##########################################################################
## Module Info
##########################################################################


__version_info__ = {"major": 5, "minor": 34, "micro": 2, "releaselevel": "final"}


##########################################################################
## Helper Functions
##########################################################################


def get_version(short=False):
    """
    Prints the version.
    """
    assert __version_info__["releaselevel"] in ("alpha", "beta", "final")
    vers = [
        "%(major)i.%(minor)i" % __version_info__,
    ]
    if __version_info__["micro"]:
        vers.append(".%(micro)i" % __version_info__)
    if __version_info__["releaselevel"] != "final" and not short:
        vers.append("%s%i" % (__version_info__["releaselevel"][0]))
    return "".join(vers)



# tests.test_transformers
# Testing package for the btrdb transformers module
#
# Author:   PingThings
# Created:  Wed Jan 02 19:26:20 2019 -0500
#
# For license information, see LICENSE.txt
# ID: test_transformers.py [] allen@pingthings.io $

"""
Testing package for the btrdb stream module
"""

##########################################################################
## Imports
##########################################################################

import os
from io import BytesIO, StringIO
from unittest.mock import Mock, PropertyMock

import numpy as np
import pytest
from pandas import DataFrame, Index, Series

from btrdb.conn import BTrDB
from btrdb.point import RawPoint, StatPoint
from btrdb.stream import Stream, StreamSet
from btrdb.transformers import *

##########################################################################
## Transformer Tests
##########################################################################


@pytest.fixture
def streamset():
    # list of mock streams
    streams = []
    for idx in range(4):
        stream = Mock(Stream)
        type(stream).collection = PropertyMock(return_value="test")
        type(stream).name = PropertyMock(return_value="stream{}".format(idx))
        stream._btrdb = Mock()
        stream._btrdb._ARROW_ENABLED = Mock(return_value=False)
        streams.append(stream)

    rows = [
        (
            None,
            RawPoint(1500000000000000000, 1.0),
            RawPoint(1500000000000000000, 1.0),
            RawPoint(1500000000000000000, 1.0),
        ),
        (
            RawPoint(1500000000100000000, 2.0),
            None,
            RawPoint(1500000000100000000, 2.0),
            RawPoint(1500000000100000000, 2.0),
        ),
        (
            None,
            RawPoint(1500000000200000000, 3.0),
            None,
            RawPoint(1500000000200000000, 3.0),
        ),
        (
            RawPoint(1500000000300000000, 4.0),
            None,
            RawPoint(1500000000300000000, 4.0),
            RawPoint(1500000000300000000, 4.0),
        ),
        (
            None,
            RawPoint(1500000000400000000, 5.0),
            RawPoint(1500000000400000000, 5.0),
            RawPoint(1500000000400000000, 5.0),
        ),
        (
            RawPoint(1500000000500000000, 6.0),
            None,
            None,
            RawPoint(1500000000500000000, 6.0),
        ),
        (
            None,
            RawPoint(1500000000600000000, 7.0),
            RawPoint(1500000000600000000, 7.0),
            RawPoint(1500000000600000000, 7.0),
        ),
        (
            RawPoint(1500000000700000000, 8.0),
            None,
            RawPoint(1500000000700000000, 8.0),
            RawPoint(1500000000700000000, 8.0),
        ),
        (
            None,
            RawPoint(1500000000800000000, 9.0),
            RawPoint(1500000000800000000, 9.0),
            RawPoint(1500000000800000000, 9.0),
        ),
        (
            RawPoint(1500000000900000000, 10.0),
            None,
            RawPoint(1500000000900000000, 10.0),
            RawPoint(1500000000900000000, 10.0),
        ),
    ]

    values = [
        [
            RawPoint(1500000000100000000, 2.0),
            RawPoint(1500000000300000000, 4.0),
            RawPoint(1500000000500000000, 6.0),
            RawPoint(1500000000700000000, 8.0),
            RawPoint(1500000000900000000, 10.0),
        ],
        [
            RawPoint(1500000000000000000, 1.0),
            RawPoint(1500000000200000000, 3.0),
            RawPoint(1500000000400000000, 5.0),
            RawPoint(1500000000600000000, 7.0),
            RawPoint(1500000000800000000, 9.0),
        ],
        [
            RawPoint(1500000000000000000, 1.0),
            RawPoint(1500000000100000000, 2.0),
            RawPoint(1500000000300000000, 4.0),
            RawPoint(1500000000400000000, 5.0),
            RawPoint(1500000000600000000, 7.0),
            RawPoint(1500000000700000000, 8.0),
            RawPoint(1500000000800000000, 9.0),
            RawPoint(1500000000900000000, 10.0),
        ],
        [
            RawPoint(1500000000000000000, 1.0),
            RawPoint(1500000000100000000, 2.0),
            RawPoint(1500000000200000000, 3.0),
            RawPoint(1500000000300000000, 4.0),
            RawPoint(1500000000400000000, 5.0),
            RawPoint(1500000000500000000, 6.0),
            RawPoint(1500000000600000000, 7.0),
            RawPoint(1500000000700000000, 8.0),
            RawPoint(1500000000800000000, 9.0),
            RawPoint(1500000000900000000, 10.0),
        ],
    ]

    obj = StreamSet(streams)
    obj.rows = Mock(return_value=rows)
    obj.values = Mock(return_value=values)
    return obj


@pytest.fixture
def statpoint_streamset():
    rows = [
        [
            StatPoint(1500000000100000000, 0, 2.0, 2.5, 10, 1),
            StatPoint(1500000000100000000, 0, 4.0, 4.5, 11, 1),
            StatPoint(1500000000100000000, 0, 6.0, 6.5, 10, 1),
            StatPoint(1500000000100000000, 0, 8.0, 8.5, 11, 2),
        ],
        [
            StatPoint(1500000000300000000, 0, 3.0, 3.5, 11, 1),
            StatPoint(1500000000300000000, 0, 5.0, 5.5, 10, 1),
            StatPoint(1500000000300000000, 0, 7.0, 7.5, 10, 1),
            StatPoint(1500000000300000000, 0, 9.0, 9.5, 11, 2),
        ],
        [
            StatPoint(1500000000500000000, 0, 4.0, 4.5, 10, 1),
            StatPoint(1500000000500000000, 0, 6.0, 6.5, 11, 1),
            StatPoint(1500000000500000000, 0, 8.0, 8.5, 10, 1),
            StatPoint(1500000000500000000, 0, 10.0, 10.5, 11, 2),
        ],
        [
            StatPoint(1500000000700000000, 0, 5.0, 5.5, 11, 1),
            StatPoint(1500000000700000000, 0, 7.0, 7.5, 10, 1),
            StatPoint(1500000000700000000, 0, 9.0, 9.5, 10, 1),
            StatPoint(1500000000700000000, 0, 11.0, 11.5, 11, 2),
        ],
    ]
    values = [
        [
            StatPoint(1500000000100000000, 0, 2.0, 2.5, 10, 1),
            StatPoint(1500000000300000000, 0, 4.0, 4.5, 11, 1),
            StatPoint(1500000000500000000, 0, 6.0, 6.5, 10, 1),
            StatPoint(1500000000700000000, 0, 8.0, 8.5, 11, 2),
        ],
        [
            StatPoint(1500000000100000000, 1, 3.0, 3.5, 11, 1),
            StatPoint(1500000000300000000, 1, 5.0, 5.5, 10, 1),
            StatPoint(1500000000500000000, 1, 7.0, 7.5, 10, 1),
            StatPoint(1500000000700000000, 1, 9.0, 9.5, 11, 2),
        ],
        [
            StatPoint(1500000000100000000, 2, 4.0, 4.5, 10, 1),
            StatPoint(1500000000300000000, 2, 6.0, 6.5, 11, 1),
            StatPoint(1500000000500000000, 2, 8.0, 8.5, 10, 1),
            StatPoint(1500000000700000000, 2, 10.0, 10.5, 11, 2),
        ],
        [
            StatPoint(1500000000100000000, 3, 5.0, 5.5, 11, 1),
            StatPoint(1500000000300000000, 3, 7.0, 7.5, 10, 1),
            StatPoint(1500000000500000000, 3, 9.0, 9.5, 10, 1),
            StatPoint(1500000000700000000, 3, 11.0, 11.5, 11, 2),
        ],
    ]
    streams = []
    for idx in range(4):
        stream = Mock(Stream)
        type(stream).collection = PropertyMock(return_value="test")
        type(stream).name = PropertyMock(return_value="stream{}".format(idx))
        stream._btrdb = Mock()
        streams.append(stream)

    obj = StreamSet(streams)
    obj.rows = Mock(return_value=rows)
    obj.values = Mock(return_value=values)
    obj.pointwidth = 20
    return obj


expected = {
    "to_dict": [
        {
            "time": 1500000000000000000,
            "test/stream0": None,
            "test/stream1": 1.0,
            "test/stream2": 1.0,
            "test/stream3": 1.0,
        },
        {
            "time": 1500000000100000000,
            "test/stream0": 2.0,
            "test/stream1": None,
            "test/stream2": 2.0,
            "test/stream3": 2.0,
        },
        {
            "time": 1500000000200000000,
            "test/stream0": None,
            "test/stream1": 3.0,
            "test/stream2": None,
            "test/stream3": 3.0,
        },
        {
            "time": 1500000000300000000,
            "test/stream0": 4.0,
            "test/stream1": None,
            "test/stream2": 4.0,
            "test/stream3": 4.0,
        },
        {
            "time": 1500000000400000000,
            "test/stream0": None,
            "test/stream1": 5.0,
            "test/stream2": 5.0,
            "test/stream3": 5.0,
        },
        {
            "time": 1500000000500000000,
            "test/stream0": 6.0,
            "test/stream1": None,
            "test/stream2": None,
            "test/stream3": 6.0,
        },
        {
            "time": 1500000000600000000,
            "test/stream0": None,
            "test/stream1": 7.0,
            "test/stream2": 7.0,
            "test/stream3": 7.0,
        },
        {
            "time": 1500000000700000000,
            "test/stream0": 8.0,
            "test/stream1": None,
            "test/stream2": 8.0,
            "test/stream3": 8.0,
        },
        {
            "time": 1500000000800000000,
            "test/stream0": None,
            "test/stream1": 9.0,
            "test/stream2": 9.0,
            "test/stream3": 9.0,
        },
        {
            "time": 1500000000900000000,
            "test/stream0": 10.0,
            "test/stream1": None,
            "test/stream2": 10.0,
            "test/stream3": 10.0,
        },
    ],
    "to_array": [
        np.array([2.0, 4.0, 6.0, 8.0, 10.0]),
        np.array([1.0, 3.0, 5.0, 7.0, 9.0]),
        np.array([1.0, 2.0, 4.0, 5.0, 7.0, 8.0, 9.0, 10.0]),
        np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]),
    ],
    "table": """               time    test/stream0    test/stream1    test/stream2    test/stream3
-------------------  --------------  --------------  --------------  --------------
1500000000000000000                               1               1               1
1500000000100000000               2                               2               2
1500000000200000000                               3                               3
1500000000300000000               4                               4               4
1500000000400000000                               5               5               5
1500000000500000000               6                                               6
1500000000600000000                               7               7               7
1500000000700000000               8                               8               8
1500000000800000000                               9               9               9
1500000000900000000              10                              10              10""",
    "series": None,
    "dataframe": None,
    "csv": """time,test/stream0,test/stream1,test/stream2,test/stream3
1500000000000000000,,1.0,1.0,1.0
1500000000100000000,2.0,,2.0,2.0
1500000000200000000,,3.0,,3.0
1500000000300000000,4.0,,4.0,4.0
1500000000400000000,,5.0,5.0,5.0
1500000000500000000,6.0,,,6.0
1500000000600000000,,7.0,7.0,7.0
1500000000700000000,8.0,,8.0,8.0
1500000000800000000,,9.0,9.0,9.0
1500000000900000000,10.0,,10.0,10.0
""",
}


class TestTransformers(object):
    ##########################################################################
    ## to_dict Tests
    ##########################################################################

    def test_to_dict(self, streamset):
        assert to_dict(streamset) == expected["to_dict"]

    def test_to_dict_statpoint(self, statpoint_streamset):
        expected = [
            {
                "time": 1500000000100000000,
                "test/stream0": 2.0,
                "test/stream1": 4.0,
                "test/stream2": 6.0,
                "test/stream3": 8.0,
            },
            {
                "time": 1500000000300000000,
                "test/stream0": 3.0,
                "test/stream1": 5.0,
                "test/stream2": 7.0,
                "test/stream3": 9.0,
            },
            {
                "time": 1500000000500000000,
                "test/stream0": 4.0,
                "test/stream1": 6.0,
                "test/stream2": 8.0,
                "test/stream3": 10.0,
            },
            {
                "time": 1500000000700000000,
                "test/stream0": 5.0,
                "test/stream1": 7.0,
                "test/stream2": 9.0,
                "test/stream3": 11.0,
            },
        ]
        assert to_dict(statpoint_streamset) == expected

    def test_to_dict_statpoint_agg_count(self, statpoint_streamset):
        expected = [
            {
                "time": 1500000000100000000,
                "test/stream0": 10,
                "test/stream1": 11,
                "test/stream2": 10,
                "test/stream3": 11,
            },
            {
                "time": 1500000000300000000,
                "test/stream0": 11,
                "test/stream1": 10,
                "test/stream2": 10,
                "test/stream3": 11,
            },
            {
                "time": 1500000000500000000,
                "test/stream0": 10,
                "test/stream1": 11,
                "test/stream2": 10,
                "test/stream3": 11,
            },
            {
                "time": 1500000000700000000,
                "test/stream0": 11,
                "test/stream1": 10,
                "test/stream2": 10,
                "test/stream3": 11,
            },
        ]
        assert to_dict(statpoint_streamset, agg="count") == expected

    def test_to_dict_statpoint_agg_all(self, statpoint_streamset):
        expected = [
            OrderedDict(
                [
                    ("time", 1500000000100000000),
                    ("test/stream0-min", 0),
                    ("test/stream0-mean", 2.0),
                    ("test/stream0-max", 2.5),
                    ("test/stream0-count", 10),
                    ("test/stream0-stddev", 1),
                    ("test/stream1-min", 0),
                    ("test/stream1-mean", 4.0),
                    ("test/stream1-max", 4.5),
                    ("test/stream1-count", 11),
                    ("test/stream1-stddev", 1),
                    ("test/stream2-min", 0),
                    ("test/stream2-mean", 6.0),
                    ("test/stream2-max", 6.5),
                    ("test/stream2-count", 10),
                    ("test/stream2-stddev", 1),
                    ("test/stream3-min", 0),
                    ("test/stream3-mean", 8.0),
                    ("test/stream3-max", 8.5),
                    ("test/stream3-count", 11),
                    ("test/stream3-stddev", 2),
                ]
            ),
            OrderedDict(
                [
                    ("time", 1500000000300000000),
                    ("test/stream0-min", 0),
                    ("test/stream0-mean", 3.0),
                    ("test/stream0-max", 3.5),
                    ("test/stream0-count", 11),
                    ("test/stream0-stddev", 1),
                    ("test/stream1-min", 0),
                    ("test/stream1-mean", 5.0),
                    ("test/stream1-max", 5.5),
                    ("test/stream1-count", 10),
                    ("test/stream1-stddev", 1),
                    ("test/stream2-min", 0),
                    ("test/stream2-mean", 7.0),
                    ("test/stream2-max", 7.5),
                    ("test/stream2-count", 10),
                    ("test/stream2-stddev", 1),
                    ("test/stream3-min", 0),
                    ("test/stream3-mean", 9.0),
                    ("test/stream3-max", 9.5),
                    ("test/stream3-count", 11),
                    ("test/stream3-stddev", 2),
                ]
            ),
            OrderedDict(
                [
                    ("time", 1500000000500000000),
                    ("test/stream0-min", 0),
                    ("test/stream0-mean", 4.0),
                    ("test/stream0-max", 4.5),
                    ("test/stream0-count", 10),
                    ("test/stream0-stddev", 1),
                    ("test/stream1-min", 0),
                    ("test/stream1-mean", 6.0),
                    ("test/stream1-max", 6.5),
                    ("test/stream1-count", 11),
                    ("test/stream1-stddev", 1),
                    ("test/stream2-min", 0),
                    ("test/stream2-mean", 8.0),
                    ("test/stream2-max", 8.5),
                    ("test/stream2-count", 10),
                    ("test/stream2-stddev", 1),
                    ("test/stream3-min", 0),
                    ("test/stream3-mean", 10.0),
                    ("test/stream3-max", 10.5),
                    ("test/stream3-count", 11),
                    ("test/stream3-stddev", 2),
                ]
            ),
            OrderedDict(
                [
                    ("time", 1500000000700000000),
                    ("test/stream0-min", 0),
                    ("test/stream0-mean", 5.0),
                    ("test/stream0-max", 5.5),
                    ("test/stream0-count", 11),
                    ("test/stream0-stddev", 1),
                    ("test/stream1-min", 0),
                    ("test/stream1-mean", 7.0),
                    ("test/stream1-max", 7.5),
                    ("test/stream1-count", 10),
                    ("test/stream1-stddev", 1),
                    ("test/stream2-min", 0),
                    ("test/stream2-mean", 9.0),
                    ("test/stream2-max", 9.5),
                    ("test/stream2-count", 10),
                    ("test/stream2-stddev", 1),
                    ("test/stream3-min", 0),
                    ("test/stream3-mean", 11.0),
                    ("test/stream3-max", 11.5),
                    ("test/stream3-count", 11),
                    ("test/stream3-stddev", 2),
                ]
            ),
        ]
        assert to_dict(statpoint_streamset, agg="all") == expected

    ##########################################################################
    ## to_array Tests
    ##########################################################################

    def test_to_array(self, streamset):
        """
        asserts to_array of rawpoint returns multidimensional ndarry of values
        """
        result = to_array(streamset)
        for idx in range(4):
            assert (result[idx] == expected["to_array"][idx]).all()

    def test_to_array_statpoint(self, statpoint_streamset):
        """
        asserts to_array of statpoint returns ndarry of requested values
        """
        result = to_array(statpoint_streamset)
        assert result.__class__ == np.ndarray
        assert (result[0] == np.array([2.0, 4.0, 6.0, 8.0])).all()
        assert (result[1] == np.array([3.0, 5.0, 7.0, 9.0])).all()
        assert (result[2] == np.array([4.0, 6.0, 8.0, 10.0])).all()
        assert (result[3] == np.array([5.0, 7.0, 9.0, 11.0])).all()

        result = to_array(statpoint_streamset, agg="min")
        assert result.__class__ == np.ndarray
        assert (
            result[0]
            == np.array(
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ]
            )
        ).all()
        assert (
            result[1]
            == np.array(
                [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                ]
            )
        ).all()
        assert (
            result[2]
            == np.array(
                [
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                ]
            )
        ).all()
        assert (
            result[3]
            == np.array(
                [
                    3.0,
                    3.0,
                    3.0,
                    3.0,
                ]
            )
        ).all()

    ##########################################################################
    ## to_series Tests
    ##########################################################################

    def test_to_series(self, streamset):
        """
        Asserts to_series produces correct results including series name
        """
        expected = []
        expected_names = [s.collection + "/" + s.name for s in streamset]
        for idx, points in enumerate(streamset.values()):
            values, times = list(zip(*[(p.value, p.time) for p in points]))
            times = Index(times, dtype="datetime64[ns]")
            expected.append(Series(values, times, name=expected_names[idx]))

        result = to_series(streamset)
        for idx in range(4):
            assert (result[idx] == expected[idx]).all()  # verify data
            assert result[idx].name == expected_names[idx]  # verify name

    def test_to_series_name_lambda(self, streamset):
        """
        assert to_series uses name lambda
        """
        result = streamset.to_series(name_callable=lambda s: s.name)
        assert [s.name for s in result] == ["stream0", "stream1", "stream2", "stream3"]

    def test_to_series_ignores_rawpoint_with_agg(self, streamset):
        """
        assert to_series ignores agg if RawPoints are used
        """
        result = streamset.to_series(agg="mean")
        assert len(result) == 4

    def test_to_series_statpoint(self, statpoint_streamset):
        """
        Asserts to_series produces correct results with statpoints
        """
        result = to_series(statpoint_streamset)
        values = [2.0, 4.0, 6.0, 8.0]
        index = [
            1500000000100000000,
            1500000000300000000,
            1500000000500000000,
            1500000000700000000,
        ]
        times = Index(index, dtype="datetime64[ns]")
        assert (result[0] == Series(values, index=times)).all()

        values = [3.0, 5.0, 7.0, 9.0]
        index = [
            1500000000100000000,
            1500000000300000000,
            1500000000500000000,
            1500000000700000000,
        ]
        times = Index(index, dtype="datetime64[ns]")
        assert (result[1] == Series(values, index=times)).all()

        result = to_series(statpoint_streamset, agg="max")
        values = [2.5, 4.5, 6.5, 8.5]
        index = [
            1500000000100000000,
            1500000000300000000,
            1500000000500000000,
            1500000000700000000,
        ]
        times = Index(index, dtype="datetime64[ns]")
        assert (result[0] == Series(values, index=times)).all()

    def test_to_series_index_type(self, streamset):
        """
        assert default index type is 'datetime64[ns]'
        """
        result = to_series(streamset)
        for series in result:
            assert series.index.dtype.name == "datetime64[ns]"

    def test_to_series_index_as_int(self, streamset):
        """
        assert datetime64_index: False produces 'int64' index
        """
        result = to_series(streamset, False)
        for series in result:
            assert series.index.dtype.name == "int64"

    def test_to_series_raises_on_agg_all(self, statpoint_streamset):
        """
        asserts to_series raises error if using "all" as agg.
        """
        with pytest.raises(AttributeError):
            statpoint_streamset.to_series(agg="all")

    ##########################################################################
    ## to_dataframe Tests
    ##########################################################################

    def test_to_dataframe(self, streamset):
        """
        assert to_dateframe works on RawPoints
        """
        columns = [
            "time",
            "test/stream0",
            "test/stream1",
            "test/stream2",
            "test/stream3",
        ]
        df = DataFrame(expected["to_dict"], columns=columns)
        df.set_index("time", inplace=True)
        assert to_dataframe(streamset).equals(df)

    def test_to_dataframe_column_issues_error(self, statpoint_streamset):
        """
        assert to_dateframe with column argument issues error
        """
        columns = ["test/cats", "test/dogs", "test/horses", "test/fishes"]
        with pytest.raises(TypeError) as unexpected_key_err:
            statpoint_streamset.to_dataframe(columns=columns)
        assert "got an unexpected keyword argument 'columns'" in str(
            unexpected_key_err.value
        )

    def test_to_dataframe_multindex(self, statpoint_streamset):
        """
        assert to_dateframe agg=all creates columned multiindex
        """
        df = statpoint_streamset.to_dataframe(agg="all")
        assert len(df.columns.levels) == 3
        assert list(df.columns.levels[0]) == ["test"]
        assert list(df.columns.levels[1]) == [
            "stream0",
            "stream1",
            "stream2",
            "stream3",
        ]
        assert list(df.columns.levels[2]) == ["count", "max", "mean", "min", "stddev"]

    def test_to_dataframe_name_lambda(self, streamset):
        """
        assert to_dateframe uses name lambda
        """
        columns = ["stream0", "stream1", "stream2", "stream3"]
        df = streamset.to_dataframe(name_callable=lambda s: s.name)

        assert len(columns) == len(df.columns)
        for idx in range(len(columns)):
            assert columns[idx] == df.columns[idx]

    def test_to_dataframe_rawpoint_with_agg(self, streamset):
        """
        assert to_dateframe ignores agg if RawPoint
        """
        df = streamset.to_dataframe(agg="all")
        assert df.shape == (10, 4)

    def test_to_dataframe_statpoints(self, statpoint_streamset):
        """
        assert to_dateframe works on StatPoints
        """
        df = statpoint_streamset.to_dataframe()

        assert df["test/stream0"].tolist() == [2.0, 3.0, 4.0, 5.0]
        assert df["test/stream1"].tolist() == [4.0, 5.0, 6.0, 7.0]
        assert df["test/stream2"].tolist() == [6.0, 7.0, 8.0, 9.0]
        assert df["test/stream3"].tolist() == [8.0, 9.0, 10.0, 11.0]

    def test_to_dataframe_agg(self, statpoint_streamset):
        """
        assert to_dateframe respects agg argument for StatPoints
        """
        df = statpoint_streamset.to_dataframe(agg="count")
        assert df["test/stream0"].tolist() == [10, 11, 10, 11]

        df = statpoint_streamset.to_dataframe(agg="min")
        assert df["test/stream3"].tolist() == [0, 0, 0, 0]

        df = statpoint_streamset.to_dataframe(agg="mean")
        assert df["test/stream0"].tolist() == [2, 3, 4, 5]

        df = statpoint_streamset.to_dataframe(agg="max")
        assert df["test/stream0"].tolist() == [2.5, 3.5, 4.5, 5.5]

        df = statpoint_streamset.to_dataframe(agg="stddev")
        assert df["test/stream0"].tolist() == [1, 1, 1, 1]

    ##########################################################################
    ## to_csv Tests
    ##########################################################################

    def test_to_csv_as_path(self, streamset, tmpdir):
        path = os.path.join(tmpdir.dirname, "to_csv_test.csv")
        streamset.to_dict = Mock(return_value=expected["to_dict"])
        to_csv(streamset, path)
        with open(path, "r") as f:
            content = f.read()
        assert content == expected["csv"]
        os.remove(path)

    def test_to_csv_as_stringio(self, streamset):
        # NOTE: There are newline issues with StringIO and csv doesnt allow
        #       BytesIO object.
        string_obj = StringIO()
        streamset.to_dict = Mock(return_value=expected["to_dict"])
        to_csv(streamset, string_obj)
        string_obj.seek(0)
        reader = csv.DictReader(string_obj)

        # convert back to dicts for assert due to newline issues with CSV
        result = [dict(row) for row in reader]
        for item in result:
            for k, v in item.items():
                if v == "":
                    item[k] = None
                elif "." in v:
                    item[k] = float(v)
                else:
                    item[k] = int(v)
        assert result == expected["to_dict"]

    def test_to_csv_raises_on_agg_all(self, statpoint_streamset):
        """
        asserts to_csv raises error if using "all" as agg.
        """
        with pytest.raises(AttributeError):
            statpoint_streamset.to_csv("tmp.txt", agg="all")

    ##########################################################################
    ## to_table Tests
    ##########################################################################

    def test_to_table(self, streamset):
        """
        asserts to_table returns formatted ASCII table
        """
        assert to_table(streamset) == expected["table"]

    def test_to_table(self, statpoint_streamset):
        """
        asserts to_table handles statpoints with requested key
        """
        expected = """               time    test/stream0    test/stream1    test/stream2    test/stream3
-------------------  --------------  --------------  --------------  --------------
1500000000100000000               2               4               6               8
1500000000300000000               3               5               7               9
1500000000500000000               4               6               8              10
1500000000700000000               5               7               9              11"""
        assert to_table(statpoint_streamset) == expected

        expected = """               time    test/stream0    test/stream1    test/stream2    test/stream3
-------------------  --------------  --------------  --------------  --------------
1500000000100000000             2.5             4.5             6.5             8.5
1500000000300000000             3.5             5.5             7.5             9.5
1500000000500000000             4.5             6.5             8.5            10.5
1500000000700000000             5.5             7.5             9.5            11.5"""
        assert to_table(statpoint_streamset, agg="max") == expected

    def test_to_table_name_callable(self, statpoint_streamset):
        """
        asserts to_table handles name_callable
        """
        expected = """               time    stream0    stream1    stream2    stream3
-------------------  ---------  ---------  ---------  ---------
1500000000100000000          2          4          6          8
1500000000300000000          3          5          7          9
1500000000500000000          4          6          8         10
1500000000700000000          5          7          9         11"""
        assert to_table(statpoint_streamset, name_callable=lambda s: s.name) == expected

    def test_to_table_raises_on_agg_all(self, statpoint_streamset):
        """
        asserts to_table raises error if using "all" as agg.
        """
        with pytest.raises(AttributeError):
            to_table(statpoint_streamset, agg="all")



# Copyright (c) 2017 Sam Kumar <samkumar@berkeley.edu>
# Copyright (c) 2017 Michael P Andersen <m.andersen@cs.berkeley.edu>
# Copyright (c) 2017 University of California, Berkeley
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of California, Berkeley nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import io
import typing
import uuid

from btrdb.exceptions import check_proto_stat, error_handler
from btrdb.grpcinterface import btrdb_pb2, btrdb_pb2_grpc
from btrdb.point import RawPoint
from btrdb.utils.general import unpack_stream_descriptor

try:
    import pyarrow as pa
except ImportError:
    pa = None

_ARROW_IMPORT_MSG = """Package `pyarrow` required, please pip install."""


class Endpoint(object):
    """Server endpoint where we make specific requests."""

    def __init__(self, channel):
        self.stub = btrdb_pb2_grpc.BTrDBStub(channel)

    @error_handler
    def rawValues(self, uu, start, end, version=0):
        params = btrdb_pb2.RawValuesParams(
            uuid=uu.bytes, start=start, end=end, versionMajor=version
        )
        for result in self.stub.RawValues(params):
            check_proto_stat(result.stat)
            yield result.values, result.versionMajor

    @error_handler
    def arrowRawValues(self, uu, start, end, version=0, schema=None):
        if pa is None:
            raise ImportError(_ARROW_IMPORT_MSG)
        templateBytes = b""
        if schema is not None:
            byte_io = io.BytesIO()
            with pa.ipc.new_stream(sink=byte_io, schema=schema) as _:
                pass
            templateBytes = byte_io.getvalue()
        params = btrdb_pb2.ArrowRawValuesParams(
            uuid=uu.bytes,
            start=start,
            end=end,
            versionMajor=version,
            templateBytes=templateBytes,
        )
        for result in self.stub.ArrowRawValues(params):
            check_proto_stat(result.stat)
            with pa.ipc.open_stream(result.arrowBytes) as reader:
                yield reader.read_all(), result.versionMajor

    @error_handler
    def arrowMultiValues(
        self, uu_list, start, end, version_list, snap_periodNS=None, schema=None
    ):
        if pa is None:
            raise ImportError(_ARROW_IMPORT_MSG)
        templateBytes = b""
        if schema is not None:
            byte_io = io.BytesIO()
            with pa.ipc.new_stream(sink=byte_io, schema=schema) as _:
                pass
            templateBytes = byte_io.getvalue()
        params = btrdb_pb2.ArrowMultiValuesParams(
            uuid=[uu.bytes for uu in uu_list],
            start=start,
            end=end,
            versionMajor=[ver for ver in version_list],
            snapPeriodNs=int(snap_periodNS),
            templateBytes=templateBytes,
        )
        for result in self.stub.ArrowMultiValues(params):
            check_proto_stat(result.stat)
            with pa.ipc.open_stream(result.arrowBytes) as reader:
                yield reader.read_all()

    @error_handler
    def arrowInsertValues(self, uu: uuid.UUID, values: pa.Table, policy: str):
        if pa is None:
            raise ImportError(_ARROW_IMPORT_MSG)
        policy_map = {
            "never": btrdb_pb2.MergePolicy.NEVER,
            "equal": btrdb_pb2.MergePolicy.EQUAL,
            "retain": btrdb_pb2.MergePolicy.RETAIN,
            "replace": btrdb_pb2.MergePolicy.REPLACE,
        }
        byte_io = io.BytesIO()
        with pa.ipc.new_stream(sink=byte_io, schema=values.schema) as writer:
            writer.write(values)
        arrow_bytes = byte_io.getvalue()
        params = btrdb_pb2.ArrowInsertParams(
            uuid=uu.bytes,
            sync=False,
            arrowBytes=arrow_bytes,
            merge_policy=policy_map[policy],
        )
        result = self.stub.ArrowInsert(params)
        check_proto_stat(result.stat)
        return result.versionMajor

    @error_handler
    def alignedWindows(self, uu, start, end, pointwidth, version=0):
        params = btrdb_pb2.AlignedWindowsParams(
            uuid=uu.bytes,
            start=start,
            end=end,
            versionMajor=version,
            pointWidth=int(pointwidth),
        )
        for result in self.stub.AlignedWindows(params):
            check_proto_stat(result.stat)
            yield result.values, result.versionMajor

    @error_handler
    def arrowAlignedWindows(self, uu, start, end, pointwidth, version=0):
        if pa is None:
            raise ImportError(_ARROW_IMPORT_MSG)
        params = btrdb_pb2.AlignedWindowsParams(
            uuid=uu.bytes,
            start=start,
            end=end,
            versionMajor=version,
            pointWidth=int(pointwidth),
        )
        for result in self.stub.ArrowAlignedWindows(params):
            check_proto_stat(result.stat)
            with pa.ipc.open_stream(result.arrowBytes) as reader:
                yield reader.read_all(), result.versionMajor

    @error_handler
    def windows(self, uu, start, end, width, depth, version=0):
        params = btrdb_pb2.WindowsParams(
            uuid=uu.bytes,
            start=start,
            end=end,
            versionMajor=version,
            width=width,
            depth=depth,
        )
        for result in self.stub.Windows(params):
            check_proto_stat(result.stat)
            yield result.values, result.versionMajor

    @error_handler
    def arrowWindows(self, uu, start, end, width, depth, version=0):
        if pa is None:
            raise ImportError(_ARROW_IMPORT_MSG)
        params = btrdb_pb2.WindowsParams(
            uuid=uu.bytes,
            start=start,
            end=end,
            versionMajor=version,
            width=width,
            depth=depth,
        )
        for result in self.stub.ArrowWindows(params):
            check_proto_stat(result.stat)
            with pa.ipc.open_stream(result.arrowBytes) as reader:
                yield reader.read_all(), result.versionMajor

    @error_handler
    def streamInfo(self, uu, omitDescriptor, omitVersion):
        params = btrdb_pb2.StreamInfoParams(
            uuid=uu.bytes, omitVersion=omitVersion, omitDescriptor=omitDescriptor
        )
        result = self.stub.StreamInfo(params)
        desc = result.descriptor
        check_proto_stat(result.stat)
        tagsanns = unpack_stream_descriptor(desc)
        return (
            desc.collection,
            desc.propertyVersion,
            tagsanns[0],
            tagsanns[1],
            result.versionMajor,
        )

    @error_handler
    def obliterate(self, uu):
        params = btrdb_pb2.ObliterateParams(uuid=uu.bytes)
        result = self.stub.Obliterate(params)
        check_proto_stat(result.stat)

    @error_handler
    def setStreamAnnotations(self, uu, expected, changes, removals):
        annkvlist = []
        for k, v in changes.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value=v)
            kv = btrdb_pb2.KeyOptValue(key=k, val=ov)
            annkvlist.append(kv)
        params = btrdb_pb2.SetStreamAnnotationsParams(
            uuid=uu.bytes,
            expectedPropertyVersion=expected,
            changes=annkvlist,
            removals=removals,
        )
        result = self.stub.SetStreamAnnotations(params)
        check_proto_stat(result.stat)

    @error_handler
    def setStreamTags(self, uu, expected, tags, collection):
        tag_data = []
        for k, v in tags.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value=v)
            kv = btrdb_pb2.KeyOptValue(key=k, val=ov)
            tag_data.append(kv)
        params = btrdb_pb2.SetStreamTagsParams(
            uuid=uu.bytes,
            expectedPropertyVersion=expected,
            tags=tag_data,
            collection=collection,
        )
        result = self.stub.SetStreamTags(params)
        check_proto_stat(result.stat)

    @error_handler
    def create(self, uu, collection, tags, annotations):
        tagkvlist = []
        for k, v in tags.items():
            kv = btrdb_pb2.KeyOptValue(key=k, val=btrdb_pb2.OptValue(value=v))
            tagkvlist.append(kv)
        annkvlist = []
        for k, v in annotations.items():
            kv = btrdb_pb2.KeyOptValue(key=k, val=btrdb_pb2.OptValue(value=v))
            annkvlist.append(kv)
        params = btrdb_pb2.CreateParams(
            uuid=uu.bytes, collection=collection, tags=tagkvlist, annotations=annkvlist
        )
        result = self.stub.Create(params)
        check_proto_stat(result.stat)

    @error_handler
    def listCollections(self, prefix):
        """
        Returns a generator for windows of collection paths matching search

        Yields
        ------
        collection paths : list[str]
        """
        params = btrdb_pb2.ListCollectionsParams(prefix=prefix)
        for msg in self.stub.ListCollections(params):
            check_proto_stat(msg.stat)
            yield msg.collections

    @error_handler
    def lookupStreams(self, collection, isCollectionPrefix, tags, annotations):
        tagkvlist = []
        for k, v in tags.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value=v)
            kv = btrdb_pb2.KeyOptValue(key=k, val=ov)
            tagkvlist.append(kv)
        annkvlist = []
        for k, v in annotations.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value=v)
            kv = btrdb_pb2.KeyOptValue(key=k, val=ov)
            annkvlist.append(kv)
        params = btrdb_pb2.LookupStreamsParams(
            collection=collection,
            isCollectionPrefix=isCollectionPrefix,
            tags=tagkvlist,
            annotations=annkvlist,
        )
        for result in self.stub.LookupStreams(params):
            check_proto_stat(result.stat)
            yield result.results

    @error_handler
    def nearest(self, uu, time, version, backward):
        params = btrdb_pb2.NearestParams(
            uuid=uu.bytes, time=time, versionMajor=version, backward=backward
        )
        result = self.stub.Nearest(params)
        check_proto_stat(result.stat)
        return result.value, result.versionMajor

    @error_handler
    def changes(self, uu, fromVersion, toVersion, resolution):
        params = btrdb_pb2.ChangesParams(
            uuid=uu.bytes,
            fromMajor=fromVersion,
            toMajor=toVersion,
            resolution=resolution,
        )
        for result in self.stub.Changes(params):
            check_proto_stat(result.stat)
            yield result.ranges, result.versionMajor

    @error_handler
    def insert(self, uu, values, policy):
        policy_map = {
            "never": btrdb_pb2.MergePolicy.NEVER,
            "equal": btrdb_pb2.MergePolicy.EQUAL,
            "retain": btrdb_pb2.MergePolicy.RETAIN,
            "replace": btrdb_pb2.MergePolicy.REPLACE,
        }
        protoValues = RawPoint.to_proto_list(values)
        params = btrdb_pb2.InsertParams(
            uuid=uu.bytes,
            sync=False,
            values=protoValues,
            merge_policy=policy_map[policy],
        )
        result = self.stub.Insert(params)
        check_proto_stat(result.stat)
        return result.versionMajor

    @error_handler
    def deleteRange(self, uu, start, end):
        params = btrdb_pb2.DeleteParams(uuid=uu.bytes, start=start, end=end)
        result = self.stub.Delete(params)
        check_proto_stat(result.stat)
        return result.versionMajor

    @error_handler
    def info(self):
        params = btrdb_pb2.InfoParams()
        result = self.stub.Info(params)
        check_proto_stat(result.stat)
        return result

    @error_handler
    def faultInject(self, typ, args):
        params = btrdb_pb2.FaultInjectParams(type=typ, params=args)
        result = self.stub.FaultInject(params)
        check_proto_stat(result.stat)
        return result.rv

    @error_handler
    def flush(self, uu):
        params = btrdb_pb2.FlushParams(uuid=uu.bytes)
        result = self.stub.Flush(params)
        check_proto_stat(result.stat)

    @error_handler
    def getMetadataUsage(self, prefix):
        params = btrdb_pb2.MetadataUsageParams(prefix=prefix)
        result = self.stub.GetMetadataUsage(params)
        check_proto_stat(result.stat)
        return result.tags, result.annotations

    @error_handler
    def generateCSV(
        self, queryType, start, end, width, depth, includeVersions, *streams
    ):
        protoStreams = [
            btrdb_pb2.StreamCSVConfig(
                version=stream[0], label=stream[1], uuid=stream[2].bytes
            )
            for stream in streams
        ]
        params = btrdb_pb2.GenerateCSVParams(
            queryType=queryType.to_proto(),
            startTime=start,
            endTime=end,
            windowSize=width,
            depth=depth,
            includeVersions=includeVersions,
            streams=protoStreams,
        )
        for result in self.stub.GenerateCSV(params):
            check_proto_stat(result.stat)
            yield result.row

    @error_handler
    def sql_query(self, stmt, params: typing.List):
        request = btrdb_pb2.SQLQueryParams(query=stmt, params=params)
        for page in self.stub.SQLQuery(request):
            check_proto_stat(page.stat)
            yield page.SQLQueryRow

    @error_handler
    def subscribe(self, update_queue):
        def updates():
            while True:
                update = update_queue.get()
                if update is None:
                    return
                (to_add, to_remove) = update
                if len(to_add) != 0:
                    yield btrdb_pb2.SubscriptionUpdate(
                        op=0, uuid=[uu.bytes for uu in to_add]
                    )
                if len(to_remove) != 0:
                    yield btrdb_pb2.SubscriptionUpdate(
                        op=1, uuid=[uu.bytes for uu in to_remove]
                    )

        for response in self.stub.Subscribe(updates()):
            check_proto_stat(response.stat)
            with pa.ipc.open_stream(response.arrowBytes) as reader:
                yield uuid.UUID(bytes=response.uuid), reader.read_all()



# tests.test_base
# Testing package for the btrdb library.
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: test_base.py [] allen@pingthings.io $

"""
Testing package for the btrdb database library.
"""

##########################################################################
## Imports
##########################################################################

import os
from unittest.mock import MagicMock, Mock, patch

import pytest

from btrdb import BTRDB_API_KEY, BTRDB_ENDPOINTS, Endpoint, __version__, connect
from btrdb.exceptions import ConnectionError

##########################################################################
## Initialization Tests
##########################################################################


class TestConnect(object):
    def setup_method(self):
        for env in ["BTRDB_ENDPOINTS", "BTRDB_PROFILE", "BTRDB_API_KEY"]:
            try:
                del os.environ[env]
            except KeyError:
                pass

    @patch("btrdb.credentials")
    def test_raises_err_with_no_inputs(self, mock_credentials):
        """
        Assert ConnectionError raised if no connection arg, ENV, profile
        """
        mock_credentials.return_value = {}
        msg = "Could not determine credentials to use."
        with pytest.raises(ConnectionError, match=msg):
            connect()

    def test_raises_err_if_both_profile_and_credentials(self):
        """
        Assert error is raised if both profile and credentials are sent
        """
        with pytest.raises(ValueError):
            connect("192.168.1.100:4410", None, "default")

    @patch("btrdb.utils.credentials.credentials_by_profile")
    @patch("btrdb.utils.credentials.credentials_by_env")
    @patch("btrdb._connect")
    def test_uses_args_over_env(
        self, mock_connect, mock_credentials_by_env, mock_credentials_by_profile
    ):
        """
        Assert uses connect args over env
        """
        mock_credentials_by_profile.return_value = {}
        mock_credentials_by_env.return_value = {"endpoints": "a", "apikey": "b"}
        connect("cat", "dog")
        mock_connect.assert_called_once_with(endpoints="cat", apikey="dog")

    @patch("btrdb.utils.credentials.credentials_by_profile")
    @patch("btrdb.utils.credentials.credentials_by_env")
    @patch("btrdb._connect")
    def test_uses_env_over_profile(
        self, mock_connect, mock_credentials_by_env, mock_credentials_by_profile
    ):
        """
        Assert connect uses env over profile info
        """
        mock_credentials_by_profile.return_value = {"endpoints": "a", "apikey": "b"}
        mock_credentials_by_env.return_value = {"endpoints": "c", "apikey": "d"}
        connect()
        mock_connect.assert_called_once_with(endpoints="c", apikey="d")

    # @patch("btrdb.utils.credentials.credentials_by_env")
    @patch("btrdb.utils.credentials.credentials_by_env")
    @patch("btrdb.utils.credentials.credentials_by_profile")
    @patch("btrdb.Connection")
    def test_connect_with_env(
        self, mock_conn, mock_credentials_by_profile, mock_credentials_by_env
    ):
        """
        Assert connect uses ENV variables
        """
        with patch(
            "btrdb.endpoint.Endpoint.info", return_value=Mock(Endpoint)
        ) as mock_ep_info:
            mock_ep_info.return_value = MagicMock()
            mock_credentials_by_profile.return_value = {}
            address = "127.0.0.1:4410"
            mock_credentials_by_env.return_value = {"endpoints": address}
            os.environ[BTRDB_ENDPOINTS] = address

            btrdb = connect()
            mock_conn.assert_called_once_with(address, apikey=None)
            mock_conn.reset_mock()

            apikey = "abcd"
            os.environ[BTRDB_API_KEY] = apikey
            mock_credentials_by_env.return_value = {
                "endpoints": address,
                "apikey": apikey,
            }
            btrdb = connect()
            mock_conn.assert_called_once_with(address, apikey=apikey)



# btrdb4
# Package stub for the btrdb4 database library.
#
# Author:   PingThings
# Created:  Fri Feb 22 11:48:05 2019 -0500
#
# For license information, see LICENSE.txt
# ID: __init__.py [] allen@pingthings.io $

"""
Package stub for the btrdb4 database library.
"""

##########################################################################
## Imports
##########################################################################

import warnings

##########################################################################
## Warning
##########################################################################

MESSAGE = (
    "The btrdb4 package is not available in this version. To install "
    "with pip use `pip install btrdb==4.*`"
)

# raise exception on ImportWarning
warnings.simplefilter("error", ImportWarning)

# warn user of version issue
warnings.warn(MESSAGE, ImportWarning)



# btrdb
# Package for the btrdb database library.
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: __init__.py [] allen@pingthings.io $

"""
Package for the btrdb database library.
"""

##########################################################################
## Imports
##########################################################################

from warnings import warn

from btrdb.conn import BTrDB, Connection
from btrdb.endpoint import Endpoint
from btrdb.exceptions import ConnectionError
from btrdb.stream import MAXIMUM_TIME, MINIMUM_TIME
from btrdb.utils.credentials import credentials, credentials_by_profile
from btrdb.utils.ray import register_serializer
from btrdb.version import get_version

##########################################################################
## Module Variables
##########################################################################

__version__ = get_version()

BTRDB_ENDPOINTS = "BTRDB_ENDPOINTS"
BTRDB_API_KEY = "BTRDB_API_KEY"
BTRDB_PROFILE = "BTRDB_PROFILE"


##########################################################################
## Functions
##########################################################################


def _connect(endpoints=None, apikey=None):
    return BTrDB(Endpoint(Connection(endpoints, apikey=apikey).channel))


def connect(conn_str=None, apikey=None, profile=None, shareable=False):
    """
    Connect to a BTrDB server.

    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. ``192.168.1.1:4411``.
        If set to None, will look in the environment variable ``$BTRDB_ENDPOINTS``
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable ``$BTRDB_API_KEY``.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        ``~/.predictivegrid/credentials.yaml``.
    shareable: bool, default=False
        Whether the connection can be "shared" in a distributed setting such
        as Ray workers. If set to True, the connection can be serialized and sent
        to other workers so that data can be retrieved in parallel; **however**, this
        is less secure because it is possible for other users of the Ray cluster to
        use your API key to fetch data.

    Returns
    -------
    db : BTrDB
        An instance of the BTrDB context to directly interact with the database.

    Examples
    --------
    This example looks for the env variables: ``BTRDB_ENDPOINTS`` and ``BTRDB_API_KEY``.

    >>> conn = btrdb.connect()
    <btrdb.conn.BTrDB at 0x...>


    Connect to the platform by looking for the relevant platform profile
    in ``${HOME}/.predictivegrid/credentials.yaml`` if the file is present.

    >>> conn = btrdb.connect(profile='test')
    <btrdb.conn.BTrDB at 0x...>

    If you provide incorrect credentials, you will get an error.

    >>> conn = btrdb.connect(conn_str="192.168.1.1:4411", apikey="NONSENSICAL_API_KEY")


    """
    # do not allow user to provide both address and profile
    if conn_str and profile:
        raise ValueError("Received both conn_str and profile arguments.")

    # check shareable flag and register custom serializer if necessary
    if shareable:
        warn(
            "a shareable connection is potentially insecure; other users of the same cluster may be able to access your API key"
        )
        register_serializer(conn_str=conn_str, apikey=apikey, profile=profile)

    # use specific profile if requested
    if profile:
        return _connect(**credentials_by_profile(profile))

    # resolve credentials using combination of arguments, env
    creds = credentials(conn_str, apikey)
    if "endpoints" in creds:
        return _connect(**creds)

    raise ConnectionError("Could not determine credentials to use.")



# btrdb.utils.credentials
# Package for interacting with the PredictiveGrid credentials file.
#
# Author:   PingThings
# Created:  Thu May 09 11:36:53 2019 -0400
#
# For license information, see LICENSE.txt
# ID: credentials.py [] allen@pingthings.io $

"""
Package for interacting with the PredictiveGrid credentials file.
"""

##########################################################################
## Imports
##########################################################################

import os
from functools import wraps

import yaml

from btrdb.exceptions import CredentialsFileNotFound, ProfileNotFound

##########################################################################
## Module Variables
##########################################################################

CONFIG_DIR = ".predictivegrid"
CREDENTIALS_FILENAME = "credentials.yaml"
CREDENTIALS_PATH = os.path.join(
    os.path.expanduser("~"), CONFIG_DIR, CREDENTIALS_FILENAME
)

##########################################################################
## Functions
##########################################################################


def filter_none(f):
    """
    decorator for removing dict items with None as value
    """

    @wraps(f)
    def inner(*args, **kwargs):
        return {k: v for k, v in f(*args, **kwargs).items() if v is not None}

    return inner


def load_credentials_from_file():
    """
    Returns a dict of the credentials file contents

    Returns
    -------
    dict
        A dictionary of profile connection information
    """
    try:
        with open(CREDENTIALS_PATH, "r") as stream:
            return yaml.safe_load(stream)
    except FileNotFoundError as exc:
        raise CredentialsFileNotFound(
            "Cound not find `{}`".format(CREDENTIALS_PATH)
        ) from exc


@filter_none
def credentials_by_profile(name=None):
    """
    Returns the BTrDB connection information (as dict) for a requested profile
    from the user's credentials file.

    Parameters
    ----------
    name: str
        The name of the profile to retrieve

    Returns
    -------
    dict
        A dictionary of the requested profile's connection information

    Raises
    ------
    CredentialsFileNotFound
        The expected credentials file `~/.predictivegrid/credentials.yaml` could not be found.

    ProfileNotFound
        The requested profile could not be found in the credentials file
    """
    if not name:
        name = os.environ.get("BTRDB_PROFILE", "default")

    # load from credentials yaml file
    creds = load_credentials_from_file()
    if name not in creds.keys():
        if name == "default":
            return {}
        raise ProfileNotFound(
            "Profile `{}` not found in credentials file.".format(name)
        )

    # rename api_key if needed and return
    fragment = creds[name].get("btrdb", {})
    if "api_key" in fragment:
        fragment["apikey"] = fragment.pop("api_key")
    return fragment


@filter_none
def credentials_by_env():
    """
    Returns the BTrDB connection information (as dict) using BTRDB_ENDPOINTS and
    BTRDB_API_KEY ENV variables.

    Returns
    -------
    dict
        A dictionary containing connection information
    """
    return {
        "endpoints": os.environ.get("BTRDB_ENDPOINTS", None),
        "apikey": os.environ.get("BTRDB_API_KEY", None),
    }


def credentials(endpoints=None, apikey=None):
    """
    Returns the BTrDB connection information (as dict) for a requested profile
    from the user's credentials file.

    Parameters
    ----------
    name: str
        The name of the profile to retrieve

    Returns
    -------
    dict
        A dictionary of the requested profile's connection information

    """
    creds = {}
    credentials_by_arg = filter_none(
        lambda: {
            "endpoints": endpoints,
            "apikey": apikey,
        }
    )
    pipeline = [credentials_by_env, credentials_by_arg]
    if os.path.exists(CREDENTIALS_PATH):
        pipeline.insert(0, credentials_by_profile)

    [creds.update(func()) for func in pipeline]
    return creds



# btrdb.utils.timez
# Conversion utilities for btrdb
#
# Author:   PingThings
# Created:  Wed Jan 02 17:00:49 2019 -0500
#
# Copyright (C) 2018 PingThings LLC
# For license information, see LICENSE.txt
#
# ID: timez.py [] allen@pingthings.io $

"""
Time related utilities
"""

##########################################################################
## Imports
##########################################################################

from datetime import datetime
from decimal import Decimal
from operator import mul

import pytz

##########################################################################
## Module Variables
##########################################################################

DATETIME_FORMATS = (
    "%Y-%m-%d %H:%M:%S.%f%z",  # most common RFC3339 nanoseconds
    "%Y-%m-%d %H:%M:%S.%f",  # expects UTC default timezone
    "%Y-%m-%dT%H:%M:%S.%fZ",  # JSON encoding, UTC timezone
    "%Y-%m-%dT%H:%M:%SZ",  # JSON encoding, UTC timezone
    "%Y-%m-%dT%H:%M:%S.%f%z",  # less common JSON-ish encoding
    "%Y-%m-%dT%H:%M:%S.%f",  # for completeness, UTC default timezone
    "%Y-%m-%d %H:%M:%S%z",  # human readable date time with TZ
    "%Y-%m-%d %H:%M:%S",  # human readable date time UTC default
    "%Y-%m-%d",  # helper to get midnight on a particular date
)


##########################################################################
## Functions
##########################################################################


def currently_as_ns():
    """
    Returns the current UTC time as nanoseconds since epoch
    """
    dt = datetime.utcnow()
    return int(dt.replace(tzinfo=pytz.utc).timestamp() * 1e9)


def ns_to_datetime(ns):
    """
    Converts nanoseconds to a naive datetime object (UTC+0)

    Parameters
    ----------
    ns : int
        nanoseconds since epoch

    Returns
    -------
    nanoseconds since epoch as a datetime object : datetime

    """
    dt = datetime.utcfromtimestamp(ns / 1e9)
    return dt.replace(tzinfo=pytz.utc)


def datetime_to_ns(dt):
    """
    Converts a datetime object to nanoseconds since epoch.  If a timezone aware
    object is received then it will be converted to UTC.

    Parameters
    ----------
    dt : datetime

    Returns
    -------
    nanoseconds : int

    """
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        aware = pytz.utc.localize(dt)
    else:
        aware = dt

    dt_utc = aware.astimezone(pytz.utc)
    return int(dt_utc.timestamp() * 1e9)


def to_nanoseconds(val):
    """
    Converts datetime, datetime64, float, str (RFC 2822) to nanoseconds.  If a
    datetime-like object is received then nanoseconds since epoch is returned.

    Parameters
    ----------
    val : datetime, datetime64, float, str
        an object to convert to nanoseconds

    Returns
    -------
    object converted to nanoseconds : int

    Notes
    ----
    The following string formats are supported for conversion.

    +--------------------------------+------------------------------------------+
    | Format String                  | Description                              |
    +================================+==========================================+
    | %Y-%m-%d %H:%M:%S.%f%z         | RFC3339 format                           |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%d %H:%M:%S.%f           | RFC3339 with UTC default timezone        |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%dT%H:%M:%S.%fZ          | JSON encoding, UTC timezone              |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%dT%H:%M:%SZ             | JSON encoding, UTC timezone, without μs  |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%dT%H:%M:%S.%f%z         | JSON-like encoding                       |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%dT%H:%M:%S.%f           | JSON-like encoding, UTC default timezone |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%d %H:%M:%S%z            | human readable date time with TZ         |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%d %H:%M:%S              | human readable date time UTC default     |
    +--------------------------------+------------------------------------------+
    | %Y-%m-%d                       | midnight at a particular date            |
    +--------------------------------+------------------------------------------+

    """
    if val is None or isinstance(val, int):
        return val

    try:
        import numpy as np

        if isinstance(val, np.datetime64):
            val = val.astype(datetime)
    except ImportError:
        pass

    if isinstance(val, str):
        # handle int as string
        if val.isdigit():
            return int(val)

        # handle datetime as string
        for format in DATETIME_FORMATS:
            try:
                val = datetime.strptime(val, format)
                break
            except ValueError:
                pass

        if isinstance(val, str):
            raise ValueError("unsupported string format, please use RFC3339")

    if isinstance(val, datetime):
        return datetime_to_ns(val)

    if isinstance(val, float):
        if val.is_integer():
            return int(val)
        else:
            raise ValueError("can only convert whole numbers to nanoseconds")

    raise TypeError("only int, float, str, datetime, and datetime64 are allowed")


def ns_delta(
    days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0, nanoseconds=0
):
    """
    Similar to ``timedelta``, ``ns_delta`` represents a span of time but as
    the total number of nanoseconds.

    Parameters
    ----------
    days : int, float, decimal.Decimal
        days (as 24 hours) to convert to nanoseconds
    hours : int, float, decimal.Decimal
        hours to convert to nanoseconds
    minutes : int, float, decimal.Decimal
        minutes to convert to nanoseconds
    seconds : int, float, decimal.Decimal
        seconds to convert to nanoseconds
    milliseconds : int, float, decimal.Decimal
        milliseconds to convert to nanoseconds
    microseconds : int, float, decimal.Decimal
        microseconds to convert to nanoseconds
    nanoseconds : int
        nanoseconds to add to the time span

    Returns
    -------
    amount of time in nanoseconds : int


    Examples
    --------
    1 minute time delta should be 60 billion nanoseconds

    >>> deltaT = ns_delta(minutes=1)
    >>> deltaT == 1 * 60 * 10**9 # 1 minute * 60 seconds * 1billion nanoseconds / second
    True
    """
    MICROSECOND = 1000
    MILLISECOND = MICROSECOND * 1000
    SECOND = MILLISECOND * 1000
    MINUTE = SECOND * 60
    HOUR = MINUTE * 60
    DAY = HOUR * 24

    if not isinstance(nanoseconds, int):
        raise TypeError("nanoseconds argument must be an integer")

    units = []
    for unit in (days, hours, minutes, seconds, milliseconds, microseconds):
        if isinstance(unit, float):
            unit = Decimal(unit)
        units.append(unit)

    factors = [DAY, HOUR, MINUTE, SECOND, MILLISECOND, MICROSECOND]
    nanoseconds += sum(map(mul, units, factors))
    return int(nanoseconds)



import os

import pyarrow as pa
import pytest

import btrdb


@pytest.fixture
def single_stream_values_arrow_schema():
    schema = pa.schema(
        [
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field("value", pa.float64(), nullable=False),
        ]
    )
    return schema


@pytest.fixture
def single_stream_windows_all_stats_arrow_schema():
    schema = pa.schema(
        [
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field("min", pa.float64(), nullable=False),
            pa.field("mean", pa.float64(), nullable=False),
            pa.field("max", pa.float64(), nullable=False),
            pa.field("count", pa.uint64(), nullable=False),
            pa.field("stddev", pa.float64(), nullable=False),
        ]
    )
    return schema


def single_stream_windows_mean_stddev_count_stats_arrow_schema():
    schema = (
        pa.schema(
            [
                pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
                pa.field("mean", pa.float64(), nullable=False),
                pa.field("count", pa.uint64(), nullable=False),
                pa.field("stddev", pa.float64(), nullable=False),
            ]
        ),
    )
    return schema


@pytest.fixture
def conn():
    if os.getenv("BTRDB_INTEGRATION_TEST_PROFILE") == None:
        pytest.skip("skipping because BTRDB_INTEGRATION_TEST_PROFILE is not set")
    conn = btrdb.connect(profile=os.getenv("BTRDB_INTEGRATION_TEST_PROFILE"))
    return conn


def _delete_collection(conn, col):
    streams = conn.streams_in_collection(col)
    for stream in streams:
        stream.obliterate()


@pytest.fixture
def tmp_collection(request, conn):
    col = "btrdb_python_integration_tests/" + request.node.name + "/tmp"
    _delete_collection(conn, col)
    yield col
    _delete_collection(conn, col)



# tests.test_point
# Testing package for the btrdb point module
#
# Author:   PingThings
# Created:  Wed Jan 02 19:26:20 2019 -0500
#
# For license information, see LICENSE.txt
# ID: test_point.py [] allen@pingthings.io $

"""
Testing package for the btrdb point module
"""

##########################################################################
## Imports
##########################################################################

import pytest

from btrdb.grpcinterface import btrdb_pb2
from btrdb.point import RawPoint, StatPoint

##########################################################################
## ProtoBuff Classes
##########################################################################

RawPointProto = btrdb_pb2.RawPoint
StatPointProto = btrdb_pb2.StatPoint

##########################################################################
## RawPoint Tests
##########################################################################


class TestRawPoint(object):
    def test_create(self):
        """
        Ensure we can create the object
        """
        point = RawPoint(1, 2)
        assert point.time == 1
        assert point.value == 2

    def test_from_proto(self):
        """
        Assert from_proto creates a correct object
        """
        proto = RawPointProto(time=1, value=10)
        point = RawPoint.from_proto(proto)
        assert point.time == proto.time
        assert point.value == proto.value

    def test_from_proto_creates_rawpoint(self):
        """
        Assert from_proto creates a new instance of RawPoint
        """
        proto = RawPointProto(time=1, value=10)
        point = RawPoint.from_proto(proto)
        assert isinstance(point, RawPoint)

    def test_from_proto_list(self):
        """
        Assert from_proto_list creates valid objects
        """
        protos = [RawPointProto(time=1, value=10), RawPointProto(time=2, value=20)]
        points = RawPoint.from_proto_list(protos)
        assert points[0].time == protos[0].time
        assert points[0].value == protos[0].value
        assert points[1].time == protos[1].time
        assert points[1].value == protos[1].value

    def test_from_proto_list_returns_list_of_rawpoint(self):
        """
        Assert from_proto_list returns list of RawPoint
        """
        protos = [RawPointProto(time=1, value=10), RawPointProto(time=2, value=20)]
        points = RawPoint.from_proto_list(protos)

        assert len(points) == 2
        for point in points:
            assert isinstance(point, RawPoint)

    def test_get_returns_correct_values(self):
        """
        Assert the __get__ returns correct values
        """
        point = RawPoint(1, 10)
        assert point[0] == 1
        assert point[1] == 10

    def test_get_raises_indexerror(self):
        """
        Assert the __get__ raises IndexError
        """
        point = RawPoint(1, 1)
        with pytest.raises(IndexError):
            point[2]
        with pytest.raises(IndexError):
            point[-1]

    def test_to_proto(self):
        """
        Assert the to_proto
        """
        point = RawPoint(1, 10)
        proto = RawPoint.to_proto(point)
        assert proto.time == 1
        assert proto.value == 10
        assert proto.__class__ == btrdb_pb2.RawPoint

    def test_to_proto_list(self):
        """
        Assert the to_proto
        """
        points = [RawPoint(1, 10), RawPoint(2, 20)]
        protos = RawPoint.to_proto_list(points)
        assert protos[0].time == 1
        assert protos[0].value == 10
        assert protos[0].__class__ == btrdb_pb2.RawPoint
        assert protos[1].time == 2
        assert protos[1].value == 20
        assert protos[1].__class__ == btrdb_pb2.RawPoint

    def test_equals(self):
        """
        Assert the __eq__ works correctly
        """
        point1 = RawPoint(1, 1)
        point2 = RawPoint(1, 1)
        point3 = RawPoint(1, 2)
        point4 = RawPoint(2, 1)

        assert point1 == point2
        assert point1 != point3
        assert point1 != point4

    def test_repr_str(self):
        """
        Assert __repr__ and __str__ work correctly
        """
        point = RawPoint(1, 10)
        expected = "RawPoint(1, 10)"

        assert str(point) == expected
        assert point.__repr__() == expected


##########################################################################
## StatPoint Tests
##########################################################################


class TestStatPoint(object):
    def test_create(self):
        """
        Ensure we can create the object
        """
        StatPoint(1, 1, 1, 1, 1, 1)

    def test_from_proto(self):
        """
        Assert from_proto creates a correct object
        """
        proto = StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6)
        point = StatPoint.from_proto(proto)
        assert point.time == proto.time
        assert point.min == proto.min
        assert point.mean == proto.mean
        assert point.max == proto.max
        assert point.count == proto.count
        assert point.stddev == proto.stddev

    def test_from_proto_creates_rawpoint(self):
        """
        Assert from_proto creates a new instance of StatPoint
        """
        proto = StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6)
        point = StatPoint.from_proto(proto)
        assert isinstance(point, StatPoint)

    def test_from_proto_list(self):
        """
        Assert from_proto_list creates valid objects
        """
        protos = [
            StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
            StatPointProto(time=11, min=12, mean=13, max=14, count=15, stddev=16),
        ]
        points = StatPoint.from_proto_list(protos)
        assert points[0].time == protos[0].time
        assert points[0].min == protos[0].min
        assert points[0].mean == protos[0].mean
        assert points[0].max == protos[0].max
        assert points[0].count == protos[0].count
        assert points[0].stddev == protos[0].stddev
        assert points[1].time == protos[1].time
        assert points[1].min == protos[1].min
        assert points[1].mean == protos[1].mean
        assert points[1].max == protos[1].max
        assert points[1].count == protos[1].count
        assert points[1].stddev == protos[1].stddev

    def test_from_proto_list_returns_list_of_statpoint(self):
        """
        Assert from_proto_list returns list of StatPoint
        """
        protos = [
            StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
            StatPointProto(time=11, min=12, mean=13, max=14, count=15, stddev=16),
        ]
        points = StatPoint.from_proto_list(protos)

        assert len(points) == 2
        for point in points:
            assert isinstance(point, StatPoint)

    def test_get_returns_correct_values(self):
        """
        Assert the __get__ returns correct values
        """
        point = StatPoint(1, 2, 3, 4, 5, 6)
        assert point[0] == 1
        assert point[1] == 2
        assert point[2] == 3
        assert point[3] == 4
        assert point[4] == 5
        assert point[5] == 6

    def test_get_raises_indexerror(self):
        """
        Assert the __get__ raises IndexError
        """
        point = StatPoint(1, 2, 3, 4, 5, 6)
        with pytest.raises(IndexError):
            point[6]
        with pytest.raises(IndexError):
            point[-1]



import concurrent
import uuid
from unittest.mock import Mock, PropertyMock, patch

import pytest

import btrdb
from btrdb.point import RawPoint
from btrdb.stream import Stream


@pytest.fixture
def stream1():
    uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
    stream = Mock(Stream)
    stream.version = Mock(return_value=11)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=10, value=1), 11))
    type(stream).collection = PropertyMock(return_value="fruits/apple")
    type(stream).name = PropertyMock(return_value="gala")
    stream.tags = Mock(return_value={"name": "gala", "unit": "volts"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "red"}, 11))
    stream._btrdb = Mock()
    stream._btrdb._executor = concurrent.futures.ThreadPoolExecutor()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=True)
    return stream


@pytest.fixture
def stream2():
    uu = uuid.UUID("17dbe387-89ea-42b6-864b-f505cdb483f5")
    stream = Mock(Stream)
    stream.version = Mock(return_value=22)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=20, value=1), 22))
    type(stream).collection = PropertyMock(return_value="fruits/orange")
    type(stream).name = PropertyMock(return_value="blood")
    stream.tags = Mock(return_value={"name": "blood", "unit": "amps"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "orange"}, 22))
    stream._btrdb = Mock()
    stream._btrdb._executor = Mock()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=True)
    return stream


class TestArrowStreams(object):
    pass
